﻿var traverse = require('ast-traverse');
var astUtils = require('sl-esprima-ast-utils');
var md5 = require('md5');
var utils = require('../utils');
var babelGenerator = require("babel-generator");
var globalErrorHandler = require('../global-error-handler');
var babylon = require("babylon");
var SourceMapHandler = require("./source-map-handler").SourceMapHandler;
var createNodeKey = require('./create-node-key').createNodeKey;
var guessMethodLocation = require('./guess-method-location').guessMethodLocation;
var guessMethodName = require('./guess-method-name').guessMethodName;
var AstNodeTypes = require("./ast-node-types");
var consoleFileLogger = require("../sl-logger.js").consoleFileLogger;
var branchKeyCreator = require('./branch-key.js');

function formatLocToArray(loc) {
    return [loc.line, loc.column];
}

// Brake the expression to pieces (every "or" will be different branch).
function findLogicalExpressionLeaves(node, leaves) {
    if (node.type === AstNodeTypes.LogicalExpression) {
        findLogicalExpressionLeaves(node.left, leaves);
        findLogicalExpressionLeaves(node.right, leaves);
    } else {
        leaves.push(node);
    }
}

// Calculate hash for function signature.
function calculateMethodNodeSignatureHash(node) {
    if (!node.params || node.params.length == 0)
        return "";

    var paramsStr = node.params.map(function (param) {
        return generateJsCodeFromAst(param);
    }).join(',');

    var hash = md5(paramsStr);
    return hash;
}

function calculateFunctionNodeHash(node) {
    var clonedNode = astUtils.clone(node);

    // Visit node in pre order and calculate body hash.
    traverse(clonedNode, {
        pre: function (node, parent, prop, idx) {
            // If current node is the root node given return (we want to calculate the body hash only).
            if (node == clonedNode)
                return;


            if (AstNodeTypes.isMethodNode(node)) {
                emptyMethodNode(node);
            }

        }
    });

    var generated = generateJsCodeFromAst(clonedNode);
    var hash = md5(generated);

    return hash;
}

/**
 * Taken from eslint - https://github.com/eslint/eslint/blob/183def6115cad6f17c82ef1c1a245eb22d0bee83/lib/eslint.js#L800
 */
function trimShebang(text) {
    return text.replace(/^#!([^\r\n]+)/, function (match, captured) {
        return "//" + captured;
    });
}

/**
 * When we have inner method we set it to empty block statement, because we want to calculate the hash and diff
 * for outer method only.
 * @param {*} methodNode
 */
function emptyMethodNode(methodNode) {
    methodNode.body = _createEmptyBlock();
    methodNode.params = [];
    methodNode.id = {
        name: '_',
        type: AstNodeTypes.Identifier
    };
}

function _createEmptyBlock() {
    return {
        type: AstNodeTypes.BlockStatement,
        body: []
    }
}

function createEmptyAlternate(node) {
// Istanbul counts empty alternates and takes the consequent's location
    var clonedNode = astUtils.clone(node);
    clonedNode.alternate = _createEmptyBlock();

    // Set alternate location to consequent location (same as istanbul).
    clonedNode.alternate.loc = node.consequent !== null ? node.consequent.loc : node.loc;
    return clonedNode;
}

// TODO: flat this logic. Roni
function addBranchesData(node, relativeFileName, absoluteFilename) {
    var branchesToAdd =[];
    switch (node.type) {

        // Those types contains some kind of "if else" structure.
        case AstNodeTypes.IfStatement:
        case AstNodeTypes.ConditionalExpression:
            var branchIndex = 0;

            // consequent mean when condition equals true.
            if (node.consequent !== null) {
                branchesToAdd.push(createBranchData(node, node.consequent, node.type, branchIndex++, relativeFileName, absoluteFilename));
            }

            // altrnate equivalent to "else".
            var clonedNode = null;
            if (node.alternate == null) {
                clonedNode = createEmptyAlternate(node);
            }
            var nodeToAdd = clonedNode || node;
            branchesToAdd.push(createBranchData(nodeToAdd, nodeToAdd.alternate, node.type, branchIndex++, relativeFileName, absoluteFilename));
            break;
        case AstNodeTypes.SwitchStatement:
            var branchIndex = 0;

            // Create branch from each case statement.
            node.cases.forEach(function (caseNode) {
                if (caseNode !== null) {
                    branchesToAdd.push(createBranchData(node, caseNode, node.type, branchIndex++, relativeFileName, absoluteFilename));
                }
            });
            break;
        case AstNodeTypes.LogicalExpression:
            var branchIndex = 0;
            var leaves = [];
            findLogicalExpressionLeaves(node, leaves);
            leaves.forEach(function (leafNode) {
                if (leafNode !== null) {
                    branchesToAdd.push(createBranchData(node, leafNode, node.type, branchIndex++, relativeFileName, absoluteFilename));
                }
            });
            break;
    }
    return branchesToAdd;
}

function emptyBranchNode(node) {
    switch (node.type) {
        case AstNodeTypes.IfStatement:
        case AstNodeTypes.ConditionalExpression:
            node.consequent = _createEmptyBlock();
            node.alternate = _createEmptyBlock();
            break;
        case AstNodeTypes.SwitchStatement:
            node.cases = [];
            break;
        case AstNodeTypes.LogicalExpression:
            node.operator = "";
            node.left = {
                type: AstNodeTypes.StringLiteral,
                value: null
            };
            node.right = {
                type: AstNodeTypes.StringLiteral,
                value: null
            };
            break;
    }
}

function createBranchData(parentNode, node, nodeType, branchIndex, relativeFileName, absoluteFileName) {
    var clonedNode = astUtils.clone(node);
    var position = formatLocToArray(node.loc.start);
    var endPosition = formatLocToArray(node.loc.end);
    //In case of AstNodeTypes.IfStatement Istanbul reports in runtime on the location of the parent node.
    var parentNodeStart = (nodeType === AstNodeTypes.ConditionalExpression || nodeType === AstNodeTypes.LogicalExpression || nodeType === AstNodeTypes.SwitchStatement) ? node.loc.start : parentNode.loc.start;
    var parentPosition = formatLocToArray(parentNodeStart)
    traverse(clonedNode, {
        pre: function (node, parent, prop, idx) {
            if (node == clonedNode) return;

            if (AstNodeTypes.isMethodNode(node)){
                emptyMethodNode(node);
            }

            // Also empty inner branches.
            emptyBranchNode(node);
        }
    });
    var generated = generateJsCodeFromAst(clonedNode);
    var hash = md5(generated);
    return {
        type: nodeType,
        index: branchIndex,
        srcData: {}, //Will be updated later in the flow.
        generatedData: {
            position: position,
            endPosition: endPosition,
            parentPosition: parentPosition,
            hash: hash,
            relativeFilename: relativeFileName,
            absoluteFilename: absoluteFileName
        }
    };
}

function replaceWithOriginalPosition(position, sourceMaps) {
    var originalPosition = sourceMaps.originalPositionFor({
        line: position[0],
        column: position[1]
    });
    if (originalPosition && originalPosition.line !== null && originalPosition.column !== null) {
        return formatLocToArray(originalPosition);
    }
    return position;
}

function updateToSourceData(handler, method, node) {
    var sourceName = handler.getOriginalName();
    if (sourceName || method.name) {
        updateMethodName(sourceName, method, node)
    }

    method.srcData.position = handler.start.asFormatArray();
    method.srcData.endPosition = handler.end.asFormatArray()
    method.srcData.displayName = handler.getOriginalName(); //Name that can be presented to the user
    //TODO: Roni, update hashes.
    method.srcData.hash = method.generatedData.hash;
    method.srcData.sigHash = method.generatedData.sigHash;
    method.srcData.relativeFilename = handler.fixedOriginalFilename;
    method.srcData.absoluteFilename = handler.originalFilename;
}

function updateToSourceDataBranch(handler, branch, node) {
    branch.srcData.position = handler.start.asFormatArray();
    branch.srcData.endPosition = handler.end.asFormatArray()
    //TODO: Roni, update hashes.
    branch.srcData.hash = branch.generatedData.hash;
    branch.srcData.relativeFilename = handler.fixedOriginalFilename;
    branch.srcData.absoluteFilename = handler.originalFilename;
}

function updateMethodName(sourceName, method, node) {
    //method.originalName = method.name;
    method.name = sourceName;
    if (!node.id)
        node.id = {
            name: sourceName
        };
    else if (!node.id.name)
        node.id.name = sourceName;
}

function generateFileSignature(filename, relativeFileName, source, sourceMaps, isBranchCoverage, opts, logger) {
    filename = utils.adjustPathSlashes(filename); //we want all transients to use unix / instead of windows \
    relativeFileName = utils.adjustPathSlashes(relativeFileName);

    var ast;

    try {
        source = trimShebang(source);

        //1. Parse the file
        ast = createAstFromFile(source, opts);

    } catch (parseException) {
        //TODO: Test
        globalErrorHandler.setLastError(parseException, "Failed to parse AST of " + filename + ". ");
        return {
            error: "Parse error (1)"
        };
    }

    var signature = astToSignature(ast, relativeFileName, sourceMaps, isBranchCoverage, filename);
    signature.hasSourceMaps = !!sourceMaps

    return signature;
}

function createSignatureDto(relativeFileName) {
    var signature = {
        methods: [],
        branches: [],
        filename: relativeFileName
    };

    return signature;
}

function generateJsCodeFromAst(ast) {
    return babelGenerator.default(ast, {
        comments: false
    }, '').code;

}

/**
 * Parses the source code and create an AST.
 * @param {The source code of the file} source
 * @param {A set of options that among the rest, control the way we parse the AST} opts
 */
function createAstFromFile(source, opts) {
    var sourceType = (opts && opts.es6Modules) ? 'module' : 'script';

    var plugins = ["*"];
    if (opts && opts.babylonPlugins) {
        plugins = plugins.concat(opts.babylonPlugins);
    }
    return babylon.parse(source, {
        // parse in strict mode and allow module declarations
        sourceType: sourceType,
        plugins: plugins
    })
}

function handleSourceData(sourceMaps, method, absoluteFilename, log, methodOrBranchNode) {
    try {
        var result = false;
        var handler = new SourceMapHandler(sourceMaps, method.generatedData.displayName, method.generatedData.position, method.generatedData.endPosition, absoluteFilename, log);
        if (!handler.shouldSkipped) {
            updateToSourceData(handler, method, methodOrBranchNode);
            result = true;
        }
    } catch (e) {
        globalErrorHandler.setLastError(e, 'Error calculating source maps:');
        log.error('Error calculating source maps:' + e);
    }
    return result;
}

/**
 * Traverse the AST and handle specific node types.
 * Information about the Node Types can be found in: https://github.com/babel/babylon/blob/master/ast/spec.md
 * @param {*} ast
 * @param {*} relativeFileName
 * @param {*} sourceMaps
 * @param {*} isBranchCoverage
 * @param {*} absoluteFilename
 */
function astToSignature(ast, relativeFileName, sourceMaps, isBranchCoverage, absoluteFilename) {
    var signature = createSignatureDto(relativeFileName);
    var log = consoleFileLogger("astToSignature");
    try {
        var nodeToParent = {};
        // print AST node types, pre-order (node first, then its children)
        traverse(ast, {
            pre: function (methodOrBranchNode, parentNode, prop, idx) {
                var key = createNodeKey(methodOrBranchNode);
                nodeToParent[key] = parentNode;


                // ObjectMethod is property function like "get", "set", "init". that uses in babylon. (in espirma its called Property).
                if (AstNodeTypes.isMethodNode(methodOrBranchNode)) {
                    var method = createMethodData(methodOrBranchNode, parentNode, nodeToParent, relativeFileName, absoluteFilename);

                    // If we have sourcemaps we need to get the location from original file.
                    if (sourceMaps) {
                       var shouldAddMethod =  handleSourceData(sourceMaps, method, absoluteFilename, log, methodOrBranchNode);
                       if(!shouldAddMethod){
                        return;
                       }
                    } else {
                        // For scripting languages in which source code is executed “as-is” we might assume generated files are identical to source. See "Top-level SL system architecture - Design ideas" by Vasili Galka.
                        method.srcData = method.generatedData;
                    }

                    // TODO: Is this relevant? we check for source position in source map handler.
                    // // Skip method that is not mapped to any "true" source file. It's compiler generated
                    // if (sourceMaps != null && method.srcData.relativeFilename == signature.filename) {
                    //     this.log.info("method: " + method.name + ". in file: " + fileSignature.filename + ". it's a generated method.")
                    //     return;
                    // }

                    method.uniqueId = [method.srcData.relativeFilename, method.srcData.position.join(',')].join('@')
                    signature.methods.push(method);
                } else if (isBranchCoverage && AstNodeTypes.isBranchNode(methodOrBranchNode)) {
                    var branches = [];

                    if (methodOrBranchNode.type !== AstNodeTypes.LogicalExpression) {
                        branches = branches.concat(addBranchesData(methodOrBranchNode, relativeFileName, absoluteFilename));
                    }

                    // If this node its not logical exp inside logical exp.
                    if (methodOrBranchNode.type == AstNodeTypes.LogicalExpression && (parentNode == undefined || parentNode.type !== AstNodeTypes.LogicalExpression)) {
                        branches = branches.concat(addBranchesData(methodOrBranchNode, relativeFileName, absoluteFilename));
                    }

                    branches.forEach(function (branchData) {
                        if (sourceMaps) {
                            var handler = new SourceMapHandler(sourceMaps, "branch", branchData.generatedData.position, branchData.generatedData.endPosition, absoluteFilename, log);
                            if (handler.shouldSkipped) {
                                branchData.srcData = branchData.generatedData;
                                return
                            }
                            updateToSourceDataBranch(handler, branchData, methodOrBranchNode);
                            branchData.srcData.parentPosition = replaceWithOriginalPosition(branchData.generatedData.parentPosition, sourceMaps);
                        } else {
                            branchData.srcData = branchData.generatedData;
                        }
                        branchData.uniqueId = branchKeyCreator(branchData.srcData.relativeFilename, branchData.srcData.parentPosition, branchData.index);

                    })

                    signature.branches = signature.branches.concat(branches);
                }


            }
        });
    } catch (traverseException1) {
        globalErrorHandler.setLastError(traverseException1, "Failed to traverse ast. Filename: " + absoluteFilename + ". Error: ");
        return {
            error: 'Parse error(2)'
        }
    }

    return signature;
}


function createMethodData(methodNode, parentNode, nodeToParent, relativeFileName, absoluteFileName) {
    //Global functions and iterators have different location and we need to get the location from the "id" in order to match the footprints.
    var location = guessMethodLocation(methodNode, nodeToParent);

    var methodHash = calculateFunctionNodeHash(methodNode);
    var methodSignatureHash = calculateMethodNodeSignatureHash(methodNode);
    var start = formatLocToArray(location.start);
    var end = formatLocToArray(methodNode.loc.end);
    var nameAndType = guessMethodName(methodNode, parentNode);
    var name = nameAndType.name; //the actual name, the guessed name, or "(Anonymous)"
    var type = nameAndType.type;
    var isAnonymous = nameAndType.isAnonymous;

    //TODO: Consult with Roni, Vasli and Cloud team about the "meta" tag.
    //In addition, I need to add tests and think how we can collect it when the code originates
    //in a different language such typescript.
    var methodData = {
        type: type,
        srcData: {}, //Will be updated later in the flow.
        generatedData: {
            position: start,
            endPosition: end,
            displayName: name, //Name that can be presented to the user
            hash: methodHash,
            sigHash: methodSignatureHash,
            relativeFilename: relativeFileName,
            absoluteFilename: absoluteFileName,
        },
        meta:
        {
            anonymous: isAnonymous,
            isAsync: methodNode.async || false,
            isComputed: methodNode.computed || false,
            isConstructor: methodNode.kind === "constructor",
            isExpression: methodNode.expression || false,
            isGenerator: methodNode.generator || false,
            isGetter: methodNode.kind === "get",
            isMethod: methodNode.kind === "method",
            isSetter: methodNode.kind === "set",
            isStatic: methodNode.static || false,
        }
    };

    return methodData;
}

module.exports.generateFileSignature = generateFileSignature;
